#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/event.h>
#include <netinet/in.h>
#include <err.h>
#include <errno.h>
#include <Message.h>
#include <masssnmp.h>
#include <fcntl.h>
#include <sys/types.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include <err.h>
#include <errno.h>
#include <sysexits.h>
#include <arpa/inet.h>
#include <sys/uio.h>
#include <unistd.h>

typedef struct task {
    struct in_addr  addr;
    ObjectName_t   *oid;
    int             request_type;
    struct task    *next;
    int             state;
    struct timeval  sent;
    ObjectName_t   *start;
    char           *community;
    int             commlen;
}               Task_struct;

typedef struct task *taskpointer;

taskpointer     beginning, task_tout = 0;
int             have_tout = 0;
int             rec_tout = 0;
int             vflag = 0;
FILE           *streamf;

#define PDU_READY     0
#define PDU_SENT      1
#define PDU_COMPLETED 2
#define PDU_TOUT      3

#define MAX_TIMESLOT  1000
#define TIMEOUT	      999999    /* TIMEOUT in us for packet to arrive */
#define DELAY_US      5000

#define PACKET_MAX    1500      /* SHOULD_CHECK the constant size buffer */
#define CONST_ID      0         /* 1191479902 */
#define SIZE_TASK     4         /* Initial Size of Task array */

#define SNMP_VERSION_2c    1
#define SNMP_PORT          161

static void usage(int);
int print_arcs(ObjectName_t *);
int add_task(int);
int check_time(struct timeval *, long, int);
int send_pdu(int, taskpointer, int);
int prepare_pdu(taskpointer, int, Message_t *);
int intree(ObjectName_t *, ObjectName_t *);
void read_pdu(int, taskpointer);

static void
usage(int ex)
{
    fprintf(stderr, "Usage:\nmassnmp [-f FILE ][-v]\n-f DebugInfoFile\n-v verbose (each increases verbosity level)\n");
    exit(ex);
}

int
print_arcs(ObjectName_t *oid)
{
    unsigned long   fixed_arcs[10];
    unsigned long  *arcs = fixed_arcs;
    int             arc_type_size = sizeof(fixed_arcs[0]);
    int             arc_slots = sizeof(fixed_arcs) / sizeof(fixed_arcs[0]);
    int             count;
    int             i;

    count = OBJECT_IDENTIFIER_get_arcs(oid, arcs,
                                       arc_type_size, arc_slots);
    if (count > arc_slots) {
        arc_slots = count;
        arcs = malloc(arc_type_size * arc_slots);
        if (!arcs)
            return (-1);
        count = OBJECT_IDENTIFIER_get_arcs(oid, arcs,
                                           arc_type_size, arc_slots);
        if (count != arc_slots)
            return (-1);
    }
    for (i = 0; i < count; i++) {
        printf(".%lu", arcs[i]);
        if (vflag > 1)
            fprintf(streamf, ".%lu", arcs[i]);
    }
    if (arcs > fixed_arcs)
        free(arcs);
    return (0);
}

int
add_task(int fd)
{
    mife_descriptor *in;
    u_char         *q;
    int             tasklen;
    babolo_parm    *bp;
    u_char         *r;
    int             res, res2;
    taskpointer     list;
    int             max_size = SIZE_TASK;
    long           *mib = 0;
    int             left_to_do = 0;
    int             i;

    if ((beginning = list = (taskpointer) malloc(sizeof(Task_struct) * max_size)) == NULL)
        errx(1, "malloc failed");

    /* Preparing the Input Stream */
    if (!(in = mife_opef(MIFE_FULL, fd))) {
        errx(1, "Reading Task failed:\n");
        exit(EX_NOINPUT);
    }
    if ((tasklen = mife_read(in, 0, 0)) < 0 || !(r = q = mife_get(in, 0))) {
        errx(1, "MIFE ERROR READING:\n");
        exit(EX_NOINPUT);
    }
    while (r < q + tasklen) {
        bp = babolo_getparm(Bpars_CRLN | Bpars_LFLN, &r, NULL, 0);
        if (!bp) {
            warnx("Input string error : \"%s\"", r);
            continue;
        }
        if (bp->argc != 3) {
            continue;
        }

        i = strlen(bp->argv[2]);
        if ((i <= 0) || (i > 32)) {
            warnx("Invalid community\n");
            continue;
        }

        if (inet_aton(bp->argv[0], &(list->addr)) == 0) {
            warnx("Not an ip %s", bp->argv[0]);
            continue;
        }

        if ((int)*(bp->argv[1]) == 46)
            bp->argv[1] += 1;
        res = OBJECT_IDENTIFIER_parse_arcs(bp->argv[1], -1, NULL, 0, NULL);
        if (res <= 0) {
            warnx("Bad oid %s", bp->argv[1]);
            continue;
        }

        if ((list->community = (char *) malloc(i * sizeof(char))) == NULL)
            errx(1, "malloc failed");

        list->commlen = i;
        bcopy(bp->argv[2], list->community, i);

        if ((mib = (long *) malloc(res * sizeof(long))) == NULL)
            errx(1, "couldn't malloc mib array");

        res2 = OBJECT_IDENTIFIER_parse_arcs(bp->argv[1], -1, mib, res, NULL);
        if (res2 != res) {
            warnx("new parsing is not equal to estimated. skipping task string");
            free(list->community);
            free(mib);
            continue;
        }

        if ((list->oid = (ObjectName_t *) malloc(sizeof(ObjectName_t))) == NULL)
            errx(1, "malloc failed");
        list->oid->size = 1;
        if ((list->oid->buf = (uint8_t *) malloc(list->oid->size * sizeof(uint8_t))) == NULL)
            errx(1, "malloc failed");
        res = OBJECT_IDENTIFIER_set_arcs(list->oid, mib, sizeof(long), res);

        if ((list->start = (ObjectName_t *) malloc(sizeof(ObjectName_t))) == NULL)
            errx(1, "malloc failed");
        list->start->size = 1;
        if ((list->start->buf = (uint8_t *) malloc(list->start->size * sizeof(uint8_t))) == NULL)
            errx(1, "malloc failed");
        res2 = OBJECT_IDENTIFIER_set_arcs(list->start, mib, sizeof(long), res2);

        if (res2 != res) {
            warnx("Some error in arcs processing. skipping task string\n");
            free(list->community);
            free(mib);
            free(list->start->buf);
            free(list->start);
            free(list->oid->buf);
            free(list->oid);
            continue;
        }

        list->request_type = PDUs_PR_get_request;
        list->state = PDU_READY;
        free(mib);

        left_to_do++;
        if (left_to_do == max_size) {
            max_size = max_size * 2;
            if (max_size / 2 != left_to_do) {
                errx(1, "Maxsize can't be more than half of max integer\n");
            }
            if ((beginning = realloc(beginning, (max_size) * sizeof(Task_struct))) == NULL)
                errx(1, "could not realloc memory");
            else {
                if (vflag)
                    warnx("reallocated to %i when reached %i", max_size, left_to_do);
                list = beginning + left_to_do - 1;
            }
        }
        list++;
    }
    return (left_to_do);
}

int
check_time(struct timeval *past, long range, int max)
{
    struct timeval *now;
    int             res;

    if ((now = (struct timeval *)malloc(sizeof(struct timeval))) == NULL)
        errx(1, "malloc failed");
    if ((gettimeofday(now, NULL)) != 0)
        errx(1, "Timer failed");
    now->tv_sec = now->tv_sec - past->tv_sec;
    now->tv_usec = now->tv_usec - past->tv_usec;
    if ((now->tv_sec < 0) || ((now->tv_usec < 0) && (now->tv_sec == 0))) {
        warnx("Timer now < Timer past. Got new timer past\n");
        res = 0;
        if ((gettimeofday(past, NULL)) != 0)
            errx(1, "Timer failed");
    } else {
        if (now->tv_sec > 2000)
            res = max;
        else {
            res = (now->tv_sec * 1000000 + now->tv_usec) / range;
        }
    }
    if (res > max)
        res = max;
    free(now);
    return (res);
}

int
send_pdu(int s, taskpointer pdu, int id)
{
    Message_t      *mesg;
    char           *outputbuf;
    struct sockaddr_in addr;
    asn_enc_rval_t  ec;
    int             result = 0, result2 = 0;

    if ((mesg = calloc(1, sizeof(Message_t))) == NULL)
        errx(1, "malloc of message failed");

    result = prepare_pdu(pdu, id, mesg);

    if (result > 0) {
        if ((outputbuf = (char *) malloc(result)) == NULL) {
            errx(1, "malloc failed");
        } else {
            ec = der_encode_to_buffer(&asn_DEF_Message, mesg, (void *)outputbuf, (size_t) result);
        }
        if (ec.encoded == result) {
            /* Create addr structure for target */
            memset(&addr, 0, sizeof(struct sockaddr_in));
            addr.sin_family = AF_INET;
            addr.sin_port = htons(SNMP_PORT);
            addr.sin_addr.s_addr = pdu->addr.s_addr;

            /* Send the Request */
            result2 = sendto(s, outputbuf, result, 0, (struct sockaddr *)&addr, sizeof(addr));
            free(outputbuf);
        }
    }

    if (pdu->request_type == PDUs_PR_get_next_request) {
        asn_DEF_VarBind.free_struct(&asn_DEF_VarBind, mesg->data.choice.get_next_request.variable_bindings.list.array[0], 0);
        asn_sequence_empty(&mesg->data.choice.get_next_request.variable_bindings.list);
    } else if (pdu->request_type == PDUs_PR_get_request) {
        asn_DEF_VarBind.free_struct(&asn_DEF_VarBind, mesg->data.choice.get_request.variable_bindings.list.array[0], 0);
        asn_sequence_empty(&mesg->data.choice.get_request.variable_bindings.list);
    }

    asn_DEF_Message.free_struct(&asn_DEF_Message, mesg, 0);

    if (result <= 0 || ec.encoded != result) {
        warnx("Encoding failure\n");
        return (-1);
    } else if (result2 != result) {
        warnx("Failed write\n");
        return (0);
    } else {
        return (1);
    }
}

int
prepare_pdu(taskpointer pdu, int id, Message_t *mes)
{
    PDU_t          *next_req;
    VarBind_t      *binder;
    asn_enc_rval_t  ec;

    /* Begin to fill in Message */
    asn_long2INTEGER(&mes->version, SNMP_VERSION_2c);

    mes->community.size = pdu->commlen;
    if ((mes->community.buf = (char *)malloc(pdu->commlen)) == NULL) {
        warnx("malloc failed");
        return (-1);
    }
    strncpy(mes->community.buf, pdu->community, pdu->commlen);
    mes->data.present = pdu->request_type;
                                
    if (mes->data.present == PDUs_PR_get_next_request) {
        next_req = &mes->data.choice.get_next_request;
    } else if (mes->data.present == PDUs_PR_get_request) {
        next_req = &mes->data.choice.get_request;
    } else {
        warnx("unsupported request type");
        return (-1);
    }
                 
    next_req->request_id = id + CONST_ID;

    asn_long2INTEGER(&next_req->error_status, 0);
    next_req->error_index = 0;
         
    next_req->variable_bindings.list.count = 0;
    next_req->variable_bindings.list.size = 0;
    if ((next_req->variable_bindings.list.array = malloc(sizeof(VarBind_t))) == NULL) {
        warnx("malloc failed");
        return (-1);
    }
                
    if ((binder = (VarBind_t *) calloc(1, sizeof(VarBind_t))) == NULL) {
        warnx("malloc of message failed");
        return (-1);
    }
    if ((binder->name.buf = (uint8_t *) malloc(pdu->oid->size)) == NULL) {
        warnx("malloc of name buffer failed");
        return (-1);
    }
    bcopy(pdu->oid->buf, binder->name.buf, pdu->oid->size);
    binder->name.size = pdu->oid->size;
    binder->value.present = VarBindtmp_PR_unSpecified;  /* The field in Request
                                                         * isn't specified */
    binder->value.choice.unSpecified = 0;

    ASN_SEQUENCE_ADD(&next_req->variable_bindings.list, binder);

    ec = der_encode(&asn_DEF_Message, mes, 0, 0);      /* To Determine the
                                                        * output buff size */
    return (ec.encoded);
}

int
intree(ObjectName_t *tree, ObjectName_t *node)
{
    int             i;
    int             min_size;

    if (tree->size > node->size)
        min_size = node->size;
    else
        min_size = tree->size;
    for (i = 0; i < min_size; i++) {
        if (tree->buf[i] != node->buf[i])
            return (1);
    }
    return (0);
}

void
read_pdu(int s, taskpointer first)
{
    Message_t      *mesg;
    ObjectName_t   *oid;
    taskpointer     received;
    int             n;
    asn_dec_rval_t  dc;
    int             request_id;
    char           *inputbuf;
    int             i, j;

    if ((inputbuf = (char *)malloc(PACKET_MAX)) == NULL)
        errx(1, "malloc failed");
    while ((n = read(s, inputbuf, PACKET_MAX)) > 0) {
        if ((mesg = (Message_t *) calloc(1, sizeof(Message_t))) == NULL)
            errx(1, "malloc failed");
        dc = asn_DEF_Message.ber_decoder(0, &asn_DEF_Message, (void **)&mesg,
                                         inputbuf, n, 0);
        if (dc.code != RC_OK) {
            warnx("Decoding failure\n");
            asn_DEF_Message.free_struct(&asn_DEF_Message, mesg, 0);
            continue;
        }
        rec_tout = 1;
        /* get the id number */
        request_id = mesg->data.choice.response.request_id - CONST_ID;
        received = first + request_id;
        if (vflag > 1) {        /* DEBUG */
            fprintf(streamf, "Got message len=%i number=%i from %s\t", n, request_id, inet_ntoa(received->addr));
            fprintf(streamf, "%i\t", *(mesg->data.choice.response.error_status.buf));
        }
        if (*(mesg->data.choice.response.error_status.buf) != 0) {
            received->state = PDU_COMPLETED;
            asn_DEF_Message.free_struct(&asn_DEF_Message, mesg, 0);
            if (vflag)
                fprintf(streamf, " Error packet received\n");
            continue;
        }
        /* Now get the oid of what we've got */
        if ((oid = (ObjectName_t *) malloc(sizeof(ObjectName_t))) == NULL)
            errx(1, "malloc failed");
        if ((oid->buf = (uint8_t *) malloc((mesg->data.choice.response.variable_bindings.list.array[0])->name.size)) == NULL)
            errx(1, "malloc failed");
        bcopy((mesg->data.choice.response.variable_bindings.list.array[0])->name.buf,
              oid->buf,
              (mesg->data.choice.response.variable_bindings.list.array[0])->name.size);
        oid->size = (mesg->data.choice.response.variable_bindings.list.array[0])->name.size;
        if ((have_tout == 1) && (received == task_tout)) {
            if (vflag)
                fprintf(streamf, "PDU_TOUT %i received. no have_tout now\n", request_id);
            have_tout = 0;
        }
        if (intree(received->start, oid)) {     /* Have reached tree's end */
            received->state = PDU_COMPLETED;
            asn_DEF_ObjectName.free_struct(&asn_DEF_ObjectName, oid, 0);
            asn_DEF_Message.free_struct(&asn_DEF_Message, mesg, 0);
            if (vflag)
                fprintf(streamf, "Tree's end\n");
            continue;
        }

        /* Proceed to next tree */
#define bindtmp mesg->data.choice.response.variable_bindings.list.array[0]->value
        if (bindtmp.present != VarBindtmp_PR_noSuchInstance || received->request_type != PDUs_PR_get_request) {
            printf("%s\t", inet_ntoa(received->addr));
            if ((i = print_arcs(oid)) != 0) {
                asn_DEF_ObjectName.free_struct(&asn_DEF_ObjectName, oid, 0);
                asn_DEF_Message.free_struct(&asn_DEF_Message, mesg, 0);
                if (vflag > 1)
                    fprintf(streamf, "\n");
                continue;
            } else {
                printf("\t");
            }
        }

        asn_DEF_ObjectName.free_struct(&asn_DEF_ObjectName, received->oid, 0);
        received->oid = oid;

        switch (bindtmp.present) {
        case VarBindtmp_PR_value:{
                switch (bindtmp.choice.value.present) {
                case ObjectSyntax_PR_simple:{
#define simtmp bindtmp.choice.value.choice.simple
                        switch (simtmp.present) {
                        case SimpleSyntax_PR_integer_value:{
                                printf("%li\n", simtmp.choice.integer_value);
                                break;
                            }
                        case SimpleSyntax_PR_string_value:{
#define pr_str simtmp.choice.string_value
                                int len = 0;
                                /* set len = 6 if buffer look like MAC address */
                                if (pr_str.size == 6 && (pr_str.buf[0] <= 31 || pr_str.buf[0] >= 128))
                                    len = 6;
                                for (j = 0; j < pr_str.size; j++) {
                                    if (!len && pr_str.buf[j] == 92)
                                        printf("\\\\");
                                    else if (!len && pr_str.buf[j] > 31 && pr_str.buf[j] < 128)
                                        printf("%c", pr_str.buf[j]);
                                    else if (len) {
                                        if (j) printf(":");
                                        printf("%.2x", pr_str.buf[j]);
                                    }
                                    else if (pr_str.buf[j] < 8)
                                        printf("\\00%o", pr_str.buf[j]);
                                    else if (pr_str.buf[j] < 64)
                                        printf("\\0%o", pr_str.buf[j]);
                                    else
                                        printf("\\%o", pr_str.buf[j]);
                                }
                                printf("\n");
#undef pr_str
                                break;
                            }
                        case SimpleSyntax_PR_objectID_value:{
                                print_arcs(&simtmp.choice.objectID_value);
                                printf("\n");
                                break;
                            }
                        default:{
                                printf("No components present\n");
                                break;
                            }
                        }
#undef simtmp
                        break;
                    }
                case ObjectSyntax_PR_application_wide:{
#define widetmp bindtmp.choice.value.choice.application_wide
                        switch (widetmp.present) {
                        case ApplicationSyntax_PR_ipAddress_value:{
                                struct in_addr  in;

                                bcopy(widetmp.choice.ipAddress_value.buf,
                                      &in,
                                      widetmp.choice.ipAddress_value.size);
                                printf("%s\n", inet_ntoa(in));
                                break;
                            }
                        case ApplicationSyntax_PR_counter_value:{
                                unsigned long   l;

                                asn_INTEGER2long(&widetmp.choice.counter_value, &l);
                                printf("%lu\n", l);
                                break;
                            }
                        case ApplicationSyntax_PR_timeticks_value:{
                                unsigned long   l;

                                asn_INTEGER2long(&widetmp.choice.timeticks_value, &l);
                                printf("%lu\n", l);     /* milliseconds */
                                break;
                            }
                        case ApplicationSyntax_PR_arbitrary_value:{
                                printf("unimplemented yet\n");
                                break;
                            }
                        case ApplicationSyntax_PR_big_counter_value:{
                                unsigned long long l;

                                asn_INTEGER2ull(&widetmp.choice.big_counter_value, &l);
                                printf("%llu\n", l);
                                break;
                            }
                        case ApplicationSyntax_PR_unsigned_integer_value:{
                                unsigned long   l;

                                asn_INTEGER2long(&widetmp.choice.unsigned_integer_value, &l);
                                printf("%lu\n", l);
                                break;
                            }
                        default:{
                                printf("No components present\n");
                            }
                        }
#undef widetmp
                        break;
                    }
                default:{
                        printf("No components present\n");
                    }
                }
                break;
            }
        case VarBindtmp_PR_unSpecified:{
                fprintf(streamf, "type unspecified\n");
                break;
            }
        case VarBindtmp_PR_noSuchObject:{
                fprintf(streamf, "no such object\n");
                break;
            }
        case VarBindtmp_PR_noSuchInstance:{
                if (received->request_type != PDUs_PR_get_request)
                    fprintf(streamf, "no such instance\n");
                break;
            }
        case VarBindtmp_PR_endOfMibView:{
                fprintf(streamf, "end of mib view\n");
                break;
            }
        default:{
                fprintf(streamf, "No components present\n");
            }
        }
#undef 	bindtmp
        received->state = PDU_READY;
        if (received->request_type == PDUs_PR_get_request)
            received->request_type = PDUs_PR_get_next_request;
        asn_DEF_Message.free_struct(&asn_DEF_Message, mesg, 0);
    }
    free(inputbuf);
}

int
main(int argc, char *argv[])
{
    int             kq;         /* Our kqueue */
    int             num_events; /* for result of kevent() call */
    int             s;          /* socket */
    struct kevent  *new_event;
    struct kevent  *events;
    int             doing = 1;  /* Flag to continue process */
    taskpointer     list, previous = 0, first;
    int             i;
    struct sockaddr_in addr;
    struct timespec *tout, *zerotout, *forcedtout;
    struct timeval *past;
    long            delay;
    long            timeout;
    int             left_to_do;
    int             timeslot = 0;
    int             q;
    int             ch;
    FILE           *fp = NULL;
    int             inda = 0;

    timeout = TIMEOUT;
    streamf = stderr;
    /* Prepare Timing Structs */
    if ((past = (struct timeval *)malloc(sizeof(struct timeval))) == NULL)
        errx(1, "malloc failed");
    delay = DELAY_US;
    if ((gettimeofday(past, NULL)) != 0)
        errx(1, "timer failed");
    /* Prepare timeouts for kqueue */
    if ((tout = (struct timespec *)malloc(sizeof(struct timespec))) == NULL)
        errx(1, "malloc failed");
    tout->tv_sec = 0;
    tout->tv_nsec = TIMEOUT * 1000;
    if ((zerotout = (struct timespec *)malloc(sizeof(struct timespec))) == NULL)
        errx(1, "malloc failed");
    zerotout->tv_sec = 0;
    zerotout->tv_nsec = 0;
    if ((forcedtout = (struct timespec *)malloc(sizeof(struct timespec))) == NULL)
        errx(1, "malloc failed");
    forcedtout->tv_sec = 0;
    forcedtout->tv_nsec = delay;
    while ((ch = getopt(argc, argv, "vf:i:")) != -1)
        switch (ch) {
        case 'v':
            vflag++;
            break;
        case 'f':
            if ((streamf = fopen(optarg, "a")) == NULL) {
                warnx("Couldn't open file for debug verbose. Changing to stderr");
                streamf = stderr;
            }
            break;
        case 'i':
            if ((fp = fopen(optarg, "r")) == NULL) {
                warnx("Couldn't open input file");
            } else {
                inda = fileno(fp);
            }
            break;
        case '?':
        default:
            usage(-1);
        }

    if (fp == NULL) {
        inda = fileno(stdin);
    }
    if (vflag)
        fprintf(streamf, "masssnmp started at %s\n", ctime(&(past->tv_sec)));

    if ((kq = kqueue()) < 0)
        errx(1, "Couldn't make kqueue");

    /* Create and bind a UDP socket */
    if ((new_event = (struct kevent *)calloc(1, sizeof(struct kevent))) == NULL)
        errx(1, "calloc failed");
    if ((events = (struct kevent *)calloc(1, sizeof(struct kevent))) == NULL)
        errx(1, "calloc failed");
    if ((s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
        errx(1, "socket failed");
    addr.sin_family = AF_INET;
    addr.sin_port = htons(0);
    addr.sin_addr.s_addr = INADDR_ANY;
    fcntl(s, F_SETFL, O_NONBLOCK);
    if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
        errx(1, "bind failed");

    /* Add descriptor to kqueue */
    EV_SET(new_event, s, EVFILT_READ, EV_ADD, 0, 0, NULL);
    if ((num_events = kevent(kq, new_event, 1, NULL, 0, 0)) < 0)
        errx(1, "adding event failed");

    /* Preparing the Task */
    left_to_do = add_task(inda);
    if (left_to_do == 0) {
        warnx("Nothing there");
        return (0);
    }
    list = beginning;
    first = list;
    for (i = 0; i < left_to_do; i++) {
        list->next = list + 1;
        list++;
    }
    list--;
    list->next = first;
    list = first;

    /* Main routine */
    while (doing) {
        /* Proceed with request */
        if ((num_events = kevent(kq, NULL, 0, events, 1, zerotout)) < 0) {
            errx(1, "waiting fot timeout event failed");
        }
        if (num_events > 0) {
            read_pdu(s, first);
        }
        if (timeslot == 0)
            timeslot += check_time(past, delay, MAX_TIMESLOT - timeslot);
        switch (list->state) {
        case PDU_READY:
            previous = list;
            if (timeslot > 0) {
                /* allowed */
                if (vflag > 1)
                    fprintf(streamf, "sending %i, %s\n", (list - first), inet_ntoa(list->addr));
                q = send_pdu(s, list, (list - first));
                if (q == 1) {
                    timeslot--;
                    if ((gettimeofday(past, NULL)) != 0)
                        errx(1, "timer failed");
                    list->sent.tv_sec = past->tv_sec;
                    list->sent.tv_usec = past->tv_usec;
                    list->state = PDU_SENT;
                    list = list->next;
                } else if (q == -1) {
                    list->state = PDU_COMPLETED;
                }
            } else {
                if ((num_events = kevent(kq, NULL, 0, events, 1, forcedtout)) < 0) {
                    errx(1, "waiting fot timeout event failed");
                }
                if (num_events > 0) {
                    read_pdu(s, first);
                }
                if (vflag)
                    warnx("No timeout available. Sleeping for forced one");
            }
            break;
        case PDU_SENT:
            previous = list;
            if (check_time(&(list->sent), timeout, 1) > 0) {
                list->state = PDU_TOUT;
            } else {
                list = list->next;
            }
            break;
        case PDU_TOUT:
            previous = list;
            if (vflag)
                fprintf(streamf, "PDU_TOUT %i have_tout=%i\n", (list - first), have_tout);
            if ((have_tout == 1) && (task_tout == list) && (rec_tout == 0)) {
                if ((num_events = kevent(kq, NULL, 0, events, 1, tout)) < 0) {
                    errx(1, "waiting fot timeout event failed");
                }
                if (num_events == 0) {
                    if (vflag)
                        fprintf(streamf, "Failed due to timeout\n");
                    warnx("Failed due to timeout\n");
                    doing = 0;
                    break;
                }
                read_pdu(s, first);
                if (rec_tout == 1) {
                    have_tout = 0;
                } else {
                    warn("In timeout packet(s) received, but all were broken\n");
                    doing = 0;
                    break;
                }
                if (vflag)
                    fprintf(streamf, "Resetting, cause received %i\n", num_events);
                break;
            } else if ((have_tout == 1) && (task_tout == list) && (rec_tout == 1)) {
                rec_tout = 0;
            }
            if (have_tout == 0) {
                have_tout = 1;
                rec_tout = 0;
                task_tout = list;
                if (vflag)
                    fprintf(streamf, "task_TOUT %i \n", (list - first));
            }
            list->state = PDU_READY;
            break;
        case PDU_COMPLETED:
            previous->next = list->next;
            left_to_do--;
            if (left_to_do == 0)
                doing = 0;
            else
                list = list->next;
            break;
        default:
            errx(1, "Unknown listtype");
        }
    }
    return (0);
}
